import abc
from PythonCoordinates.coordinates.coordinate_representations import SphericalCoordinate
from PythonCoordinates.measurables.physical_quantities import Length


class INoiseSource:
    @abc.abstractmethod
    def predict(self, location: SphericalCoordinate):
        """
        This function will utilize the information within the class to predict the acoustic metric at the specific
        emission location on the surface of the definition.

        :param location:
            This is the emission location. It is defined on the surface of a sphere inscribed around the source.

        :returns:
            A specific value, or set of values, at the specific location.
        """

        raise NotImplementedError()

    def __init__(self):
        """
        A constructor for the generic interface to the acoustic source descriptions
        """
        pass



import numpy
import numpy as np
from scipy.interpolate import SmoothSphereBivariateSpline
from scipy.stats import pearsonr
from sklearn.metrics import mean_squared_error
import matplotlib.pyplot as plt
import matplotlib as mpl
from mpl_toolkits.mplot3d import Axes3D

from physical_sources.acoustic.sphere_source import ISphereSource, SphericalFitMethods
from PythonCoordinates.measurables.physical_quantities import Length, Angle
from PythonCoordinates.coordinates.coordinate_representations import SphericalCoordinate


def plot_fit_data_2d(locations=None, metric=None, source=None, metric_name=None):
    cm = plt.cm.get_cmap('jet')
    fig = plt.figure()
    plt.subplot(projection='aitoff')
    plt.grid()
    plt.yticks(
        [5 / 12 * np.pi, 4 / 12 * np.pi, 3 / 12 * np.pi, 2 / 12 * np.pi, 1 / 12 * np.pi, 0, -1 / 12 * np.pi,
         -2 / 12 * np.pi, -3 / 12 * np.pi, -4 / 12 * np.pi, -5 / 12 * np.pi],
        labels=['15', '30', '45', '60', '75', '90', '105', '120', '135', '150', '165']
    )
    plt.text(0, -0 + np.pi / 2, '+z')
    plt.text(0, -np.pi + np.pi / 2, '-z')
    plt.text(0, -np.pi / 2 + np.pi / 2, '+x')
    plt.text(np.pi, -np.pi / 2 + np.pi / 2, '-x')
    plt.text(-np.pi, -np.pi / 2 + np.pi / 2, '-x')
    plt.text(np.pi / 2, -np.pi / 2 + np.pi / 2, '+y')
    plt.text(-np.pi / 2, -np.pi / 2 + np.pi / 2, '-y')
    plt.xlabel('azimuth ($\phi$)')
    plt.ylabel('polar ($\Theta$)')

    if metric is not None:
        vmin = np.min(metric)
        vmax = np.max(metric)
    elif source is not None:
        if not isinstance(source, GridSphereSource):
            raise TypeError("Input source must be of type 'GridSphereSource'")
        vmin = source.metric_grid.min()
        vmax = source.metric_grid.max()
    else:
        raise ValueError('Either the input pair locations and metric or the input source must not equal None.')

    if source is not None:
        theta_regular = []
        for i in range(len(source.polar)):
            theta_regular.append(-source.polar[i].radians + np.pi / 2)
        phi_regular = []
        for i in range(len(source.azimuth)):
            phi_regular.append(source.azimuth[i].radians)
        plt.pcolormesh(phi_regular, theta_regular, source.metric_grid, cmap=cm, vmin=vmin, vmax=vmax)

    if locations is not None:

        normalize_180_180(locations)

        theta = []
        phi = []
        for i in range(len(locations)):
            theta.append(-locations[i].polar.radians + np.pi / 2)
            phi.append(locations[i].azimuthal.radians)
        plt.scatter(phi, theta, c=metric, cmap=cm, vmin=vmin, vmax=vmax, edgecolors='k')

    plt.colorbar(label=metric_name)

    return fig


# TODO Fix and implement 3D Sphere Plotting
def plot_fit_data_3d(locations=None, metric=None, source = None, metric_name=None):
    # fig = plt.figure()
    # ax = plt.axes(projection='3d')
    #
    # if metric is not None:
    #     vmin = np.max(metric)
    #     vmax = np.min(metric)
    # elif source is not None:
    #     if not isinstance(source, GridSphereSource):
    #         raise TypeError("Input source must be of type 'GridSphereSource'")
    #     vmin = source.metric_grid.min()
    #     vmax = source.metric_grid.max()
    # else:
    #     raise ValueError('Either the input pair locations and metric or the input source must not equal None.')
    # norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
    #
    # if source is not None:
    #     x = np.empty([len(source.polar), len(source.azimuth) + 1])
    #     y = np.empty([len(source.polar), len(source.azimuth) + 1])
    #     z = np.empty([len(source.polar), len(source.azimuth) + 1])
    #     for i in range(len(source.polar)):
    #         for j in range(len(source.azimuth)):
    #             location = SphericalCoordinate(r=source.radius, polar=source.polar[i], azimuthal=source.azimuth[j])
    #             location = location.to_cartesian()
    #             x[i, j] = location.x.meters
    #             y[i, j] = location.y.meters
    #             z[i, j] = location.z.meters
    #         location = SphericalCoordinate(r=source.radius, polar=source.polar[i], azimuthal=source.azimuth[1])
    #         location = location.to_cartesian()
    #         x[i, -1] = location.x.meters
    #         y[i, -1] = location.y.meters
    #         z[i, -1] = location.z.meters
    #     metric_grid_wrapped = np.concatenate((source.metric_grid, source.metric_grid[:, 1, None]), axis=1)
    #
    #     ax.plot_surface(x, y, z, facecolors=plt.cm.jet(norm(metric_grid_wrapped)))
    #     m = plt.cm.ScalarMappable(cmap=plt.cm.jet, norm=norm)
    #     m.set_array([])
    #     plt.colorbar(m)
    #     ax.set_xlabel('x (m)')
    #     ax.set_ylabel('y (m)')
    #     ax.set_zlabel('z (m)')
    #     # plt.show()
    #
    # return fig
    pass


def normalize_180_180(locations):
    for i in range(len(locations)):
        while locations[i].azimuthal.degrees > 180:
            locations[i] = SphericalCoordinate(
                locations[i].r,
                locations[i].polar,
                Angle(locations[i].azimuthal.degrees - 360, unit=Angle.Units.Degrees)
            )
        while locations[i].azimuthal.degrees < -180:
            locations[i] = SphericalCoordinate(
                locations[i].r,
                locations[i].polar,
                Angle(locations[i].azimuthal.degrees + 360, unit=Angle.Units.Degrees)
            )


class GridSphereSource(ISphereSource):
    """
    This class defines any metric on the surface of a sphere of fixed radius, with regular angular
    spacing in the polar (theta) and azimuthal (phi) conventional spherical coordinates.

    :param polar: Iterable[Angle]
        List, Tuple, or 'numpy.ndarray' of polar (theta) angle values, regularly spaced in the range [0, 180]
    :param azimuth: Iterable[Angle]
        List, Tuple, or 'numpy.ndarray' of polar (theta) angle values, regularly spaced in the range [-180, 180]
    :param radius: Length
        A single value of type 'Length' denoting the fixed sphere radius
    :param metric_grid: numpy.ndarray
        A 2-dimensional 'numpy.ndarray' of size len(polar) x len(azimuth)
    """

    def __init__(self, polar=None, azimuth=None, radius: Length = None, metric_grid: numpy.ndarray = np.empty([0, 0]),
                 corr_coef: float = None, mse: float = None):
        super().__init__()

        self._polar = polar
        if self._polar is not None:
            self._theta = self._calculate_angles_degrees(self._polar)
        else:
            self._theta = None

        self._azimuth = azimuth
        if self._azimuth is not None:
            self._phi = self._calculate_angles_degrees(self._azimuth)
        else:
            self._phi = None

        self._radius = radius
        if self._radius is not None:
            self._check_radius(self._radius)

        self._metric_grid = metric_grid
        self._check_metric_grid(self._metric_grid, self.polar, self.azimuth)

        self._corr_coef = corr_coef
        self._mse = mse

    @property
    def polar(self):
        return self._polar

    @polar.setter
    def polar(self, values):

        self._polar = values
        self._theta = self._calculate_angles_degrees(self._polar)

    @property
    def azimuth(self):
        return self._azimuth

    @azimuth.setter
    def azimuth(self, values):
        self._azimuth = values
        self._phi = self._calculate_angles_degrees(self.azimuth)

    @property
    def radius(self):
        if self._radius is not None:
            self._check_radius(self._radius)
        return self._radius

    @radius.setter
    def radius(self, value):
        self._check_radius(value)
        self._radius = value

    @property
    def metric_grid(self):
        if self._metric_grid.any():
            self._check_metric_grid(self._metric_grid, self.polar, self.azimuth)
        return self._metric_grid

    @metric_grid.setter
    def metric_grid(self, array):
        self._metric_grid = array

    @property
    def mse(self):
        return self._mse

    @property
    def corr_coef(self):
        return self._corr_coef

    @staticmethod
    def fit(locations, metric, order: int = 0, dpolar: Angle = Angle(5.0, unit=Angle.Units.Degrees),
            dazimuth: Angle = Angle(5.0, unit=Angle.Units.Degrees),
            method=SphericalFitMethods.SmoothSphereBivariateSpline,
            std=1e-2):
        """
        This function will return a GridSphereSource object representing any metric by fitting the values in
        metric sampled at any set of SphericalCoordinate locations on a sphere of fixed radius.

        :param locations: Iterable[SphericalCoordinate]
            This is a collection of the known locations on the spherical surface. The radius values must be equal
            for all locations.
        :param metric: Iterable[float]
            A single or double dimensioned array that will be used to fit the regular grid values.
        :param order: int
            The order of fitting method that will be accomplished
        :param dpolar: Angle
            The desired regular polar angle spacing in degrees
        :param dazimuth: Angle
            The desired regular azimuthal angle spacing in degrees
        :param method: SphericalFitMethods
            An Enumeration of available methods for fitting metric to a spherical surface
        :param std: float
            The input parameter which controls the interplay between the smoothness of the resulting curve
            and the quality of the approximation of the data. Increasing std leads to smoother fits, and in
            the limit of a very large std, the curve degenerates into a single best-fit polynomial. A good
            choice might be around the approximate standard deviation of the data.
        :return: GridSphereSource
            A GridSphereSource object containing estimates of the metric values sampled at regular angular
            intervals over the surface of a sphere of the same radius as the input locations
        :raises ValueError: if the radius values r are not equal at all locations
        """

        # TODO - Alan implement check and warning if input points overlap in location

        GridSphereSource._check_polar_angles(locations)

        radius = GridSphereSource._get_radius(locations)
        theta = np.empty(len(locations))
        phi = np.empty(len(locations))
        for i in range(len(locations)):
            theta[i] = locations[i].polar.radians
            phi[i] = locations[i].azimuthal.normalized.radians

        grid_interpolation = SmoothSphereBivariateSpline(theta, phi, metric, s=len(locations) * std ** 2)
        metric_predicted = grid_interpolation(theta, phi, grid=False)

        theta_regular = np.arange(0., np.pi * 1.0001, step=dpolar.radians)
        phi_regular = np.arange(0, 2 * np.pi, step=dazimuth.radians)
        metric_grid = grid_interpolation(theta_regular, phi_regular)

        # Shift data to -180 <= phi <= +180 convention
        phi_regular[phi_regular < 0] += 2 * np.pi
        metric_grid = np.concatenate((
            metric_grid[:, [i for i in range(len(phi_regular)) if phi_regular[i] >= np.pi]],
            metric_grid[:, [i for i in range(len(phi_regular)) if phi_regular[i] < np.pi]],
            metric_grid[:, np.where(phi_regular >= np.pi)[0][0], None]
        ), axis=1)
        phi_regular = np.concatenate((
            phi_regular[[i for i in range(len(phi_regular)) if phi_regular[i] >= np.pi]] - 2 * np.pi,
            phi_regular[[i for i in range(len(phi_regular)) if phi_regular[i] < np.pi]],
            phi_regular[np.where(phi_regular >= np.pi)[0][0], None]
        ))

        theta_grid = np.empty(len(theta_regular), dtype=Angle)
        for i in range(len(theta_regular)):
            theta_grid[i] = Angle(theta_regular[i], unit=Angle.Units.Radians)
        phi_grid = np.empty(len(phi_regular), dtype=Angle)
        for i in range(len(phi_regular)):
            phi_grid[i] = Angle(phi_regular[i], unit=Angle.Units.Radians)

        return GridSphereSource(
            polar=theta_grid,
            azimuth=phi_grid,
            radius=radius,
            metric_grid=metric_grid,
            corr_coef=pearsonr(metric, metric_predicted).statistic,
            mse=mean_squared_error(metric, metric_predicted)
        )

    def predict(self, location):
        pass

    @staticmethod
    def parse_xml(root):
        pass

    def make_xml(self, root):
        pass

    @staticmethod
    def _calculate_angles_degrees(angles):
        """
        Calculates angles as a 'numpy.ndarray' in degrees from an iterable of type 'Angle' objects
        """
        degrees = np.empty(shape=[len(angles)])
        for i in range(len(degrees)):
            if not isinstance(angles[i], Angle):
                raise TypeError("Elements of input iterable must be of type 'Angle'")
            degrees[i] = angles[i].degrees
        return degrees

    @staticmethod
    def _check_radius(radius):
        if not isinstance(radius, Length):
            raise TypeError("radius must be of type 'Length'")

    @staticmethod
    def _get_radius(locations):
        for i in range(len(locations)):
            if locations[i].r != locations[0].r:
                raise ValueError("All locations must be given as 'SphericalCoordinate' objects with the same radius r")
        return locations[0].r

    @staticmethod
    def _check_polar_angles(locations):
        for a in locations:
            if not 0 <= a.polar.radians <= np.pi:
                raise ValueError('polar angles must lie within the interval [0, pi]')

    @staticmethod
    def _check_metric_grid(metric_grid, polar, azimuth):
        if not isinstance(metric_grid, numpy.ndarray):
            raise TypeError("metric_grid must be of type 'numpy.ndarray'")
        if len(metric_grid.shape) != 2:
            raise ValueError("metric_grid must be a 2-dimensional array")
        if metric_grid.any():
            if metric_grid.shape[0] != len(polar):
                raise ValueError("The length of the first dimension of metric_grid must equal len(polar)")
            if metric_grid.shape[1] != len(azimuth):
                raise ValueError("The length of the second dimension of metric_grid must equal len(azimuth)")


class GridSphereSourceTemp(ISphereSource):
    def __init__(self,
                 grid_sphere: GridSphereSource = None,
                 corr_coeff: float = None,
                 mse: float = None
                 ):
        super().__init__()
        self._corr_coeff = None
        self._mse = None


class Weapon(ISphereSource):

    @staticmethod
    def fit(x, y, order: int = 0):
        """
        :param x: Numpy.ndarray of SphericalCoordinate objects
            This is a collection ...
        :param y: Numpy.ndarray or Dataframe
            A single or double ...
        :param order: int
            The maximum...
        """

        # series =

        # Import data from standard file

        # Loop through the second dimenstion of the y values (i.e. metric columns)

        # source = GridSphereSource.fit()

        return "Object of type Weapon"

    def predict(self, location):
        return "all metrics at a single location"

    @staticmethod
    def parse_xml(root):
        pass

    def save(self):
        pass



from physical_sources.acoustic.point_source import FlightPointSource, GroundPointSource, AircraftType


class FlightNoiseFileDatabase:
    def __init__(self, filename):
        """
        This constructor will load the entire noise database from the flight01.dat file that is passed to the
        constructor
        :param filename: str - the path to the NOISEFILE database which we want to load
        """
        self.filename = filename
        self.sources = []
        with open(filename, 'r') as f:
            contents = f.readlines()
        line_index = 0
        while line_index < len(contents):
            src_data = FlightPointSource()
            try:
                line_index = src_data._load_data(contents, line_index)
            except Exception as eee:
                raise ValueError('An incorrect formatted NOISEFILE entry found at index {}: {}'.format(
                    line_index, str(eee)))
            self.sources.append(src_data)

    def get_sources(self, id: int, ac_type: AircraftType = AircraftType.Military) -> list:
        """
        This function will extract all the elements of the database that possess the specific aircraft_id that is passed
        to the function.

        :param id:
            the integer id for the aircraft that we want to extract

        :returns:
            A list of the aircraft with the specific id that was provided to the function
        """

        sources = [x for x in self.sources if (x.aircraft_id == id) & (x.vehicle_type == ac_type)]

        return sources

class GroundNoiseFileDatabase:
    def __init__(self, path):
        """
        This will read a file with multiple NOISEFILE_Static_AcousticSource definition in a single file and place the
        elements into an internal array
        """

        self.filename = path
        self.sources = []

        with open(self.filename, 'rt') as f:
            contents = f.readlines()

        line_index = 0

        while line_index < len(contents):
            src_data = GroundPointSource()
            try:
                line_index = src_data._load_data(contents, line_index)
            except Exception as eee:
                raise ValueError("An incorrect formatted NOISEFILE entry found at index {}: {}".format(
                    line_index, str(eee)
                ))

            self.sources.append(src_data)


import abc
import copy
from enum import Enum
import numpy as np
from PythonCoordinates.measurables.physical_quantities import Length, Angle, Speed, Temperature, Humidity, Pressure
from PythonCoordinates.coordinates.coordinate_representations import SphericalCoordinate
from physical_sources.acoustic.inoisesource import INoiseSource


class AircraftType(Enum):
    """
    Within the NOISEFILE source description there is an ability to represent the acoustic emissions from both military
    and civilian aircraft. This is an enumeration to define which type of aircraft is represented by a specific source
    description.
    """
    Military = 0, 'Military'
    Civilian = 1, 'Civilan'


class DataCollectionType(Enum):
    """
    The NOISEFILE dataset permits the definition of acoustic source emissions through estimating the acoustic levels.
    The desired method is to measure the source levels. This determines what type of data is being represented in the
    source.
    """
    Measured = 0, 'Measured'
    Estimated = 1, 'Estimated'


class InterpolationCode(Enum):
    """
    Inside the Omega10 program (which reads the NOISEFILE Dataset) there are three methods available for interpolation
    of the acoustic levels. The first is fixed and will return an error if the user attempts to interpolat this levels
    as part of the calculation of new levels. The parallel method assumes that there is a linear relationship between
    the data, but this point defines an offset point parallel to that linear relationship. The final method is variable
    which defines the linear relationship between all points marked with the variable label.
    """
    Fixed = 0, 'Fixed'
    Parallel = 1, 'Parallel'
    Variable = 2, 'Variable'


class PowerSetting:
    """
    This class contains the information that is present in the source definition that relates to the power setting of
    the aircraft. It contains an average, minimum, and maximum value for this description and the associated units.
    """
    def __init__(self):
        self._value = 0.0
        self._units = ''
        self._upper_limit = 0.0
        self._lower_limit = 0.0

    @property
    def power_setting_value(self) -> float:
        return self._value

    @power_setting_value.setter
    def power_setting_value(self, value):
        self._value = value

    @property
    def units(self) -> str:
        return self._units

    @units.setter
    def units(self, value):
        self._units = value

    @property
    def lower_limit(self) -> float:
        return self._lower_limit

    @lower_limit.setter
    def lower_limit(self, value):
        self._lower_limit = value

    @property
    def upper_limit(self) -> float:
        return self._upper_limit

    @upper_limit.setter
    def upper_limit(self, value):
        self._vupper_limit = value


class IPointSource(INoiseSource):
    """
    Older models of acoustic source
    """
    def __init__(self):
        super().__init__()

        self._vehicle_type = None
        self._interpolation_code = None
        self._id = None
        self._operational_power_code = None
        self._reference_distance = None
        self._reference_temperature = None
        self._reference_humidity = None
        self._source = None
        self._engine_name = None
        self._engine_count = None
        self._collection_date = None
        self._power_settings = []
        self._power_setting_description = None

    @abc.abstractmethod
    def predict(self, location: SphericalCoordinate):
        raise NotImplementedError

    @property
    def vehicle_type(self) -> AircraftType:
        return self._vehicle_type

    @property
    def interpolation_code(self) -> InterpolationCode:
        return self._interpolation_code

    @property
    def aircraft_id(self) -> int:
        return self._id

    @property
    def operational_power_code(self) -> int:
        return self._operational_power_code

    @property
    def reference_distance(self):
        return self._reference_distance

    @property
    def reference_temperature(self):
        return self._reference_temperature

    @property
    def reference_humidity(self):
        return self._reference_humidity

    @property
    def source(self):
        return self._source

    @property
    def engine_name(self):
        return self._engine_name

    @property
    def engine_count(self):
        return self._engine_count

    @property
    def collection_date(self):
        return self._collection_date

    @property
    def data_collection_type(self) -> DataCollectionType:
        return self._data_collection_type

    @property
    def power_settings(self):
        return self._power_settings

    @staticmethod
    def source_type():
        return 'Unknown'


    @property
    def power_setting_description(self):
        return self._power_setting_description


class FlightPointSource(IPointSource):

    def __init__(self, filename=None):
        super().__init__()
        self._microphone_count = None
        self._directivity_angle = None
        self._perceived_noise_level = None
        self._tone_corrected_perceived_noise_level = None
        self._la = None
        self._lat = None
        self._effective_perceived_noise_level = None
        self._sound_exposure_level = None
        self._tone_corrected_sound_exposure_level = None
        self._tone_correction = None
        self._operational_type_code = None
        self._drag_configuration = None
        self._reference_speed = None
        self._power_setting_description = None
        self._sound_pressure_levels = []
        self._frequency_resolution = 3
        self._minimum_frequency_band = 10
        self._maximum_frequency_band = 40

        if filename is not None:
            self.filename = filename
            f = open(self.filename)
            contents = f.readlines()
            f.close()
            self._load_data(contents, 0)



    @property
    def frequency_resolution(self):
        return self._frequency_resolution

    @property
    def minimum_frequency_band(self):
        return self._minimum_frequency_band

    @property
    def maximum_frequency_band(self):
        return self._maximum_frequency_band

    @property
    def number_microphone_in_average(self):
        return self._microphone_count

    @property
    def directivity_angle(self):
        return self._directivity_angle

    @property
    def perceived_noise_level(self):
        return self._perceived_noise_level

    @property
    def tone_corrected_pnl(self):
        return self._tone_corrected_perceived_noise_level

    @property
    def a_weighted_level(self):
        return self._la

    @property
    def tone_corrected_a_weighted_level(self):
        return self._lat

    @property
    def sound_exposure_level(self):
        return self._sound_exposure_level

    @property
    def tone_corrected_sel(self):
        return self._tone_corrected_sound_exposure_level

    @property
    def effective_perceived_noise_level(self):
        return self._effective_perceived_noise_level

    @property
    def tone_correction(self):
        return self._tone_correction

    @property
    def operational_type_code(self):
        return self._operational_type_code

    @property
    def drag_configuration(self):
        return self._drag_configuration

    @property
    def reference_speed(self):
        return self._reference_speed

    def predict(self, location):
        return self._sound_pressure_levels

    @staticmethod
    def source_type():
        return 'FLIGHT'

    def write_data(self, file):
        """
        This function will write the data from the class into a formatted file that is the same format as the current
        representation of the NOISEFILE dataset.

        :param file:
            This is the file writer that receives the bytes to write to the file.
        """
        if self.vehicle_type == AircraftType.Military:
            record1 = 'MILITARY FM{:5d}{:2d}{:1d}'.format(self.aircraft_id, self.operational_power_code,
                                                          self.operational_type_code)
        else:
            record1 = 'CIVILIAN FC{:5d}{:2d}{:1d}'.format(self.aircraft_id, self.operational_power_code,
                                                          self.operational_type_code)

        if self.interpolation_code == InterpolationCode.Fixed:
            record1 = record1 + 'F'
        elif self.interpolation_code == InterpolationCode.Parallel:
            record1 = record1 + 'P'
        elif self.interpolation_code == InterpolationCode.Variable:
            record1 = record1 + 'V'
        other1 = '{:<20}{:<20}{:>2} {:<25}'.format(self.name, self.engine_name, self.engine_count,
                                                   self.drag_configuration)
        if self.data_collection_type == DataCollectionType.Measured:
            other2 = 'MEASURED  '
        else:
            other2 = 'ESTIMATED '
        other1 = other1 + other2
        record2 = '{}{:<12}{:>11}{:>5} FT{:>5} KTS  {:>3} F  {:>3} PCT'.format(
            other1, self.source, self.collection_date, int(self.reference_distance.feet),
            int(self.reference_speed.knots), int(self.reference_temperature.fahrenheit),
            int(self.reference_humidity.percentage))
        record3 = '{:<20}'.format(self.power_setting_description)
        for setting in self.power_settings:
            record3 = record3 + '{:>9} {:<10} {:>9}{:>9}'.format('{:.2f}'.format(setting.value), setting.units,
                                                                 '{:.2f}'.format(setting.lower_limit),
                                                                 '{:.2f}'.format(setting.upper_limit))

        record4 = '{:>4}{:>6}{:>6}{:>6}{:>6}{:>6}{:>6}{:>6}{:>6}{:>6}'.format(self.number_microphone_in_average,
                                                                              self.directivity_angle.degrees,
                                                                              self.perceived_noise_level,
                                                                              self.tone_corrected_pnl,
                                                                              self.a_weighted_level,
                                                                              self.tone_corrected_a_weighted_level,
                                                                              self.effective_perceived_noise_level,
                                                                              self.sound_exposure_level,
                                                                              self.tone_corrected_sel,
                                                                              self.tone_correction)

        record5 = ''
        for level in self._sound_pressure_levels:
            record5 = record5 + '{:>4}'.format(int(level * 10))

        file.write(record1 + '\n')
        file.write(record2 + '\n')
        file.write(record3 + '\n')
        file.write(record4 + '\n')
        file.write(record5 + '\n')

    def _load_data(self, contents, start_index: int = 0):
        """
        This function loads the information for this description from the contents of a file. In the case where this may
        represent multiple point source definitions in a single file, the second parameter tells the function where to
        begin looking for the start of the source definition.
        :param contents:
            This list of rows from a file that contain the source definitions
        :param start_index:
            An index within the contents list/array for where we will start processing the information for this
            definition.
        """

        record1 = contents[start_index].rstrip()
        start_index += 1
        record2 = contents[start_index].rstrip()
        start_index += 1
        record3 = contents[start_index].rstrip()
        start_index += 1
        record4 = contents[start_index].rstrip()
        start_index += 1
        record5 = contents[start_index].rstrip()
        start_index += 1
        try:
            if record1[0:8] == 'MILITARY':
                self._vehicle_type = AircraftType.Military
            else:
                self._vehicle_type = AircraftType.Civilian
            self._id = int(record1[11:16])
            self._operational_power_code = int(record1[16:18])
            self._operational_type_code = int(record1[18:19])
            interp_code = record1[19:20]
            if interp_code == 'F':
                self._interpolation_code = InterpolationCode.Fixed
            elif interp_code == 'P':
                self._interpolation_code = InterpolationCode.Parallel
            elif interp_code == 'V':
                self._interpolation_code = InterpolationCode.Variable
            self.name = record2[0:20].strip()
            self._engine_name = record2[20:40].strip()
            self._engine_count = int(record2[40:42])
            self._drag_configuration = record2[43:43 + 25].strip()
            if record2[68:78].strip() == 'MEASURED':
                self._data_collection_type = DataCollectionType.Measured
            else:
                self._data_collection_type = DataCollectionType.Estimated
            self._source = record2[78:78 + 12].strip()
            self._collection_date = record2[90:90 + 11]
            self._reference_distance = Length(float(record2[101:101 + 5]), Length.Units.Feet)
            self._reference_speed = Speed(float(record2[109:109 + 5]), Speed.Units.Knots)
            self._reference_temperature = Temperature(float(record2[120:120 + 3]), Temperature.Units.Fahrenheit)
            self._reference_humidity = Humidity(float(record2[127:127 + 3]), Humidity.Units.Percentage)
            if len(record3) > 0:
                self._power_setting_description = record3[0:20].strip()
                record3 = record3[20:]
                while len(record3) > 0:
                    temp = PowerSetting()
                    temp.value = float(record3[0:9].strip())
                    temp.units = record3[10:10 + 10].strip()
                    temp.lower_limit = float(record3[20:20 + 9].strip())
                    temp.upper_limit = float(record3[30:30 + 9].strip())
                    self._power_settings.append(temp)
                    record3 = record3[39:]
            self._microphone_count = int(record4[0:4].strip())
            self._directivity_angle = Angle(float(record4[4:4 + 6].strip()))
            self._perceived_noise_level = float(record4[10:10 + 6].strip())
            self._tone_corrected_perceived_noise_level = float(record4[16:16 + 6].strip())
            self._la = float(record4[22:22 + 6].strip())
            self._lat = float(record4[28:28 + 6].strip())
            self._effective_perceived_noise_level = float(record4[34:34 + 6].strip())
            self._sound_exposure_level = float(record4[40:40 + 6].strip())
            self._tone_corrected_sound_exposure_level = float(record4[46:46 + 6].strip())
            self._tone_correction = float(record4[52:52 + 6].strip())

            for index in range(0, len(record5), 4):
                self._sound_pressure_levels.append(float(record5[index:index + 4].strip()) / 10.0)
        except Exception as e:
            raise ValueError('An error ocurred in the NOISEFILE formatted file at index = ' + str(
                start_index) + ' - error message: ' + str(e))
        return start_index


class GroundPointSource(IPointSource):
    def __init__(self, filename=None):

        super().__init__()
        self.noise_suppression_system = ''
        self._reference_pressure = Pressure()

        if filename is None:
            self.polar_sound_pressure_levels = {}
        else:
            self.filename = filename
            self.polar_sound_pressure_levels = {}
            with open(filename, 'r') as file:
                contents = file.readlines()
                self._load_data(contents, 0)

    @property
    def reference_pressure(self):
        return self._reference_pressure

    @property
    def angle_count(self) -> int:
        return len(self.polar_sound_pressure_levels)

    @staticmethod
    def source_type():
        return 'GROUND'

    def _load_data(self, contents, start_index):
        record1 = contents[start_index].rstrip()
        start_index += 1
        try:
            #   Record or line 1
            if record1[0:8] == 'MILITARY':
                self._vehicle_type = AircraftType.Military
            else:
                self._vehicle_type = AircraftType.Civilian
            self._id = int(record1[11:11 + 5])
            self._operational_power_code = int(record1[16:16 + 2])
            interpolation_code = record1[18:18 + 1]
            if interpolation_code == 'F':
                self._interpolation_code = InterpolationCode.Fixed
            elif interpolation_code == 'P':
                self._interpolation_code = InterpolationCode.Parallel
            elif interpolation_code == 'V':
                self._interpolation_code = InterpolationCode.Variable

            #   Record 2
            record2 = contents[start_index].rstrip()
            start_index += 1
            self.name = record2[0:20].strip()
            self._engine_name = record2[20:20 + 20].strip()
            self.noise_suppression_system = record2[40:40 + 14].strip()
            self._engine_count = int(record2[54:54 + 2])
            if 'MEASURED' in record2[57:57 + 10]:
                self._data_collection_type = DataCollectionType.Measured
            else:
                self._data_collection_type = DataCollectionType.Estimated
            self._source = record2[67:67 + 12].strip()
            self._collection_date = record2[79:79 + 11]

            #   Record 3
            record3 = contents[start_index].rstrip()
            start_index += 1
            self._power_setting_description = record3[0:20].strip()
            for i in range(0, 3, 1):
                substring = record3[20 + (i * 20): 20 + (i * 20) + 21]
                try:
                    power_setting_value = float(substring[0:9])
                    power_setting_units = substring[10:10 + 10].strip()
                    power_setting = PowerSetting()
                    power_setting.value = power_setting_value
                    power_setting.units = power_setting_units
                    self._power_settings.append(power_setting)
                except Exception as ee:
                    print('FAILURE: {}'.format(str(ee)))
                    break

            self._reference_distance = Length(float(record3[80:80 + 5]), Length.Units.Feet)
            self._reference_speed = Speed()
            self._reference_temperature = Temperature(float(record3[89:89 + 3]), Temperature.Units.Fahrenheit)
            self._reference_pressure = Pressure(float(record3[103:103 + 6]), Pressure.Units.InchesMercury)
            self._reference_humidity = Humidity(float(record3[95:95 + 3]))

            self.polar_sound_pressure_levels = np.zeros((19, 31))
            start_index += 1
            for band_index in range(10, 41, 1):
                data_record = contents[start_index].rstrip()
                start_index += 1
                temp_elements = data_record.split(' ')
                elements = [x for x in temp_elements if x != '']
                n = 1
                for angle_index in range(19):
                    self.polar_sound_pressure_levels[angle_index, band_index - 10] = float(elements[n]) / 10.0
                    n += 1
        except Exception as ee:
            raise ValueError('{} occurred in the NOISEFILE file at index = {}'.format(str(ee), start_index))
        return start_index

    def write_data(self, writer):
        if self.vehicle_type == AircraftType.Military:
            record = 'MILITARY FM{:5d}{:2d}'.format(self.aircraft_id, self.operational_power_code)
        else:
            record = 'CIVILIAN FC{:5d}{:2d}'.format(self.aircraft_id, self.operational_power_code)

        if self.interpolation_code == InterpolationCode.Fixed:
            record = record + 'F'
        elif self.interpolation_code == InterpolationCode.Parallel:
            record = record + 'P'
        elif self.interpolation_code == InterpolationCode.Variable:
            record = record + 'V'
        writer.write(record + '\n')

        record = '{:<20}{:<20}{:<14}{:>2} '.format(self.name, self.engine_name, self.noise_suppression_system,
                                                   self.engine_count)
        if self.data_collection_type == DataCollectionType.Measured:
            record = record + 'MEASURED  '
        else:
            record = record + 'ESTIMATED '

        record = record + '{:<12}{:>11} Single Engine Data'.format(self.source, self.collection_date)

        writer.write(record + '\n')
        record = '{:<20}'.format(self.power_setting_description)
        for setting in self.power_settings:
            record = record + '{:>9} {:<10}'.format('{:.2f}'.format(setting.value), setting.units)
        while len(record) < 80:
            record = record + ' '

        record = record + '{:>5} FT {:>3} F {:>3} PCT {:>6} IN HG'.format(int(self.reference_distance.feet),
                                                                          int(self.reference_temperature.fahrenheit),
                                                                          int(self.reference_humidity.percentage),
                                                                          self.reference_pressure.inches_mercury)
        writer.write(record + '\n')
        record = ' BAND  '
        for angle in range(0, 181, 10):
            record = record + '{:>5}'.format(angle)
        writer.write(record + '\n')

        for band_index in range(10, 41, 1):
            record = '  {:>2}   '.format(band_index)
            for angle_key in range(19):
                record = record + '{:>5}'.format(int(self.polar_sound_pressure_levels[angle_key, band_index - 10] * 10))

            writer.write(record + '\n')

    def predict(self, location: SphericalCoordinate) -> list:
        if isinstance(location, SphericalCoordinate):
            return self.__sound_pressure_level__(location.azimuthal)
        else:
            raise ValueError("You must use the PythonCoordinates module to define the location on the source")

    def __sound_pressure_level__(self, azimuthal) -> list:
        if isinstance(azimuthal, Angle):
            angle = azimuthal.normalized.degrees
            if angle > 180:
                angle = 360 - angle
            degree_angle_key = int(np.floor(angle / 10))
            return self.polar_sound_pressure_levels[degree_angle_key]
        else:
            raise ValueError("You must use the PythonCoordinates module to define the location on the source")


class InterpolatedPointSource(IPointSource):
    """
    This class provides a method of representing the noise power distance curve that is used by NoiseMap to determine
    the acoustic levels as a function of power setting. The power settings are described as individual IPointNoise
    sources. The interpolation is then defined in accordance with the Omega 10 interpolation schema.
    """

    def __init__(self, a: list):
        """
        This constructs the collection of sources and provides the methods for interpolating the levels as a function
        of the desired power setting
        """

        super().__init__()

        self._sources = copy.deepcopy(a)

        self._desired_power_setting = a[0].power_setting[0]

    @property
    def desired_power_setting(self) -> PowerSetting:
        return self._desired_power_setting

    @desired_power_setting.setter
    def desired_power_setting(self, value: PowerSetting):
        self._desired_power_setting = value

    @property
    def sources(self):
        return self._sources

    def predict(self, location: SphericalCoordinate):
        """
        This function will take the value in the desired_power_setting function and determine the interpolated value at
        this location in accordance with the NoiseMap program called Omega10.
        :param location:
            The spherical location were we want to evaluate the source levels for the desired_power_setting
        :returns:
            A list or array of the sound pressure levels at this location and desired_power_setting
        """

        raise NotImplementedError





from PythonCoordinates.measurables.physical_quantities import Speed, Angle, Length
from PythonCoordinates.measurables.Measurable import InvalidUnitOfMeasureException
from PythonCoordinates.coordinates.coordinate_representations import CartesianCoordinate, SphericalCoordinate
from pytimbre.spectral.fractional_octave_band import FractionalOctaveBandTools as fob
import numpy as np
from physical_sources.acoustic.inoisesource import INoiseSource
import abc


class IAnalyticSource(INoiseSource):
    """
    This Interface base class defines the collection of information that is required for the analytic methods of
    determining value of an acoustic metric at a specific spherical location.
    """

    @abc.abstrcatmethod
    def predict(self, location: SphericalCoordinate):
        """
        This function employs the internal representation of the coefficients of the analytic function.

        :param location:
            This is the location on a sphere inscribed around the source where we want to evaluate the acoustic metric.
        """
        raise NotImplementedError()


class IdealDipoleX(IAnalyticSource):
    """
    This class implements the equation from Dr. Frank Mobley's dissertation that defines an acoustic dipole with an
    arbitrary rotation about the z-axis. It provides methods to alter the strength and frequency of the dipole so that
    a series of sound pressure levels can be obtained for any spherical angle.
    """

    def predict(self, location: SphericalCoordinate):
        """
        Given a location on the surface of the directivity pattern, this will determine the sound pressure leve at each
        of the frequencies within the frequency list.

        :param location:
            SphericalCoordinate - the first location of the point that we want to process
        :returns: np.ndarray - a collection of sound pressure levels for the location given
        """

        spl = np.zeros((len(self._frequency), ))

        if not isinstance(location, SphericalCoordinate):
            raise InvalidUnitOfMeasureException("You must provide the location of the point in SphericalCoordinate")

        for i in range(len(self._frequency)):
            wavelength = self._sound_speed.meters_per_second / self._frequency[i]
            k = 1 / wavelength
            w = 2 * np.pi * self._frequency[i]

            c_pressure = -complex(0, self._amplitude * k * self._separation / self.reference_distance.meters)
            c_pressure *= complex(np.cos(w * self._time - k * self.reference_distance.meters),
                                  np.sin(w * self._time - k * self.reference_distance.meters))
            c_pressure *= np.cos(location.polar) * np.sin(location.azimuthal)
            pressure = abs(c_pressure)

            spl[i] = 20 * np.log10(pressure / 20e-6)

        return spl

    def __init__(self, dipole_separation: float = 0.001, amplitude: float = 0.1,
                 sound_speed: Speed = Speed.ref_speed_of_sound(), frequency: np.ndarray = fob.tob_frequencies()):
        """
        This constructor defines the initial conditions for the dipole class

        :param dipole_separation: float
            the distance between the two monopoles that form the dipole
        :param amplitude: float
            the strength of both monopoles
        :param sound_speed: Speed
            This is the nominal speed of sound for the determination of the wave number
        """

        self._separation = dipole_separation
        self._amplitude = amplitude
        self._sound_speed = sound_speed
        self._frequency = frequency
        self.name = "Dr. Mobley's Ideal Dipole (X-axis orientation)"
        self.minimum_frequency_band = self._frequency[0]
        self.maximum_frequency_band = self._frequency[-1]
        self._time = 0

    @property
    def calculation_time(self):
        return self._time

    @calculation_time.setter
    def calculation_time(self, value):
        self._time = value

